/*
 * Copyright (c) 2012-2013, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ======== MainDsp.c ========
 *  Platform: OMAPL138_linux_elf
 */

/* xdctools header files */
#include <xdc/std.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/System.h>

/* package header files */
#include <ti/ipc/Ipc.h>
#include <ti/ipc/MultiProc.h>
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

/* local header files */
#include "AppDsp.h"
#include "SvrDsp.h"

#define PRI_HIGH        5
#define PRI_MED         3

/* private functions */
static Void MainDsp_appTskFxn(UArg arg0, UArg arg1);
static Void MainDsp_svrTskFxn(UArg arg0, UArg arg1);
static Void MainDsp_done(Void);


/*
 *  ======== main ========
 */
Int main(Int argc, Char* argv[])
{
    Error_Block     eb;
    Task_Params     taskParams;

    System_printf("--> main:\n");

    /* must initialize the error block before using it */
    Error_init(&eb);

    System_printf("main: ipc ready\n");

    /* create server thread */
    Task_Params_init(&taskParams);
    taskParams.instance->name = "Svr";
    taskParams.priority = PRI_HIGH;
    taskParams.arg0 = (UArg)argc;
    taskParams.arg1 = (UArg)argv;
    taskParams.stackSize = 0x600;
    Task_create(MainDsp_svrTskFxn, &taskParams, &eb);

    if (Error_check(&eb)) {
        System_abort("main: failed to create server thread");
    }

    /* create application thread */
    Task_Params_init(&taskParams);
    taskParams.instance->name = "App";
    taskParams.priority = PRI_MED;
    taskParams.arg0 = (UArg)argc;
    taskParams.arg1 = (UArg)argv;
    taskParams.stackSize = 0x600;
    Task_create(MainDsp_appTskFxn, &taskParams, &eb);

    if (Error_check(&eb)) {
        System_abort("main: failed to create application thread");
    }

    /* start scheduler, this never returns */
    BIOS_start();

    /* should never get here */
    System_printf("<-- main:");
    return (0);
}

/*
 *  ======== MainDsp_appTskFxn ========
 */
Void MainDsp_appTskFxn(UArg arg0, UArg arg1)
{
    Int status = 0;

    System_printf("--> MainDsp_appTskFxn:\n");

    /* application setup phase */
    status = AppDsp_setup();

    if (status < 0) {
        goto leave;
    }

    /* application execute phase */
    status = AppDsp_run();

    if (status < 0) {
        goto leave;
    }

    /* application shutdown phase */
    status = AppDsp_destroy();

    if (status < 0) {
        goto leave;
    }

leave:
    System_printf("<-- MainDsp_appTskFxn: %d\n",
            (IArg)status);

    /* ccs debug hook */
    MainDsp_done();
}

/*
 *  ======== MainDsp_svrTskFxn ========
 */
Void MainDsp_svrTskFxn(UArg arg0, UArg arg1)
{
    Int status = 0;

    System_printf("--> MainDsp_svrTskFxn:\n");

    /* server setup phase */
    status = SvrDsp_setup();

    if (status < 0) {
        goto leave;
    }

    /* server execute phase */
    status = SvrDsp_run();

    if (status < 0) {
        goto leave;
    }

    /* server shutdown phase */
    status = SvrDsp_destroy();

    if (status < 0) {
        goto leave;
    }

leave:
    System_printf("<-- MainDsp_svrTskFxn: %d\n",
            status);
}

/*
 *  ======== MainDsp_done ========
 */
Void MainDsp_done(Void)
{
    System_printf("MainDsp_done:\n");
}
